/* 
 *	Copyright (C) Chia-chen Kuo - April 2001
 *
 *  This file is part of DVD2AVIMFC, a free MPEG-2 decoder
 *	
 *  DVD2AVIMFC is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2, or (at your option)
 *  any later version.
 *   
 *  DVD2AVIMFC is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 *  GNU General Public License for more details.
 *   
 *  You should have received a copy of the GNU General Public License
 *  along with GNU Make; see the file COPYING.  If not, write to
 *  the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA. 
 *
 */

#include "global.h"
#include "getbit.h"
#include "AC3Dec\ac3.h"
#include "WaveOut.h"

extern int SectorHit;

// Pointer to a function that handles decoding subpictures. 
void (*SubpicDecoder)(unsigned char* Buffer) = 0;

#define LOCATE												\
while (Rdptr >= (Rdbfr + BUFFER_SIZE))						\
{															\
	Refill_Buffer();										\
	Rdptr -= BUFFER_SIZE;									\
}

#define DEMUX_SUBPIC														\
while (Packet_Length > 0)													\
{																			\
	/* Send the data to whoever will handle subpic decoding */				\
	if (Packet_Length+Rdptr > BUFFER_SIZE+Rdbfr)							\
	{																		\
		Packet_Length -= BUFFER_SIZE+Rdbfr-Rdptr;							\
	    Refill_Buffer();													\
		Rdptr = Rdbfr;														\
	}																		\
	else																	\
	{																		\
		if(SubpicDecoder) SubpicDecoder(Rdbfr);								\
		Rdptr += Packet_Length;												\
		Packet_Length = 0;													\
	}																		\
}																			\

#define DECODE_AC3																\
{																				\
	size = 0;																	\
	while (Packet_Length > 0)													\
	{																			\
		if (Packet_Length+Rdptr > BUFFER_SIZE+Rdbfr)							\
		{																		\
			size = ac3_decode_data(Rdptr, BUFFER_SIZE+Rdbfr-Rdptr, size);		\
			Packet_Length -= BUFFER_SIZE+Rdbfr-Rdptr;							\
			Refill_Buffer();													\
			Rdptr = Rdbfr;														\
		}																		\
		else																	\
		{																		\
			size = ac3_decode_data(Rdptr, Packet_Length, size);					\
			Rdptr += Packet_Length;												\
			Packet_Length = 0;													\
		}																		\
	}																			\
}

#define DEMUX_AC3															\
while (Packet_Length > 0)													\
{																			\
	if (Packet_Length+Rdptr > BUFFER_SIZE+Rdbfr)							\
	{																		\
		if (Dry == false)													\
			fwrite(Rdptr, BUFFER_SIZE+Rdbfr-Rdptr, 1, ac3[This_Track].file);\
		Packet_Length -= BUFFER_SIZE+Rdbfr-Rdptr;							\
	    Refill_Buffer();													\
		Rdptr = Rdbfr;														\
	}																		\
	else																	\
	{																		\
		if (Dry == false)													\
			fwrite(Rdptr, Packet_Length, 1, ac3[This_Track].file);			\
		Rdptr += Packet_Length;												\
		Packet_Length = 0;													\
	}																		\
}

#define DEMUX_PCM														\
{																		\
	unsigned char tmp;													\
	size = 0;															\
	while (Packet_Length > 0)											\
	{																	\
		if (Packet_Length+Rdptr > BUFFER_SIZE+Rdbfr)					\
		{																\
			memcpy(PCM_Buffer+size, Rdptr, BUFFER_SIZE+Rdbfr-Rdptr);	\
			size += BUFFER_SIZE+Rdbfr-Rdptr;							\
			Packet_Length -= BUFFER_SIZE+Rdbfr-Rdptr;					\
		    Refill_Buffer();											\
			Rdptr = Rdbfr;												\
		}																\
		else															\
		{																\
			memcpy(PCM_Buffer+size, Rdptr, Packet_Length);				\
			Rdptr += Packet_Length;										\
			size += Packet_Length;										\
			Packet_Length = 0;											\
		}																\
	}																	\
	for (i=0; i<size; i+=2)												\
	{																	\
		tmp = PCM_Buffer[i];											\
		PCM_Buffer[i] = PCM_Buffer[i+1];								\
		PCM_Buffer[i+1] = tmp;											\
	}																	\
}

#define DEMUX_MPA															\
while (Packet_Length > 0)													\
{																			\
	if (Packet_Length+Rdptr > BUFFER_SIZE+Rdbfr)							\
	{																		\
		if (Dry == false)													\
			fwrite(Rdptr, BUFFER_SIZE+Rdbfr-Rdptr, 1, mpa[MPA_Track].file);	\
		Packet_Length -= BUFFER_SIZE+Rdbfr-Rdptr;							\
		Refill_Buffer();													\
		Rdptr = Rdbfr;														\
	}																		\
	else																	\
	{																		\
		if (Dry == false)													\
			fwrite(Rdptr, Packet_Length, 1, mpa[MPA_Track].file);			\
		Rdptr += Packet_Length;												\
		Packet_Length = 0;													\
	}																		\
}

#define DEMUX_DTS															\
while (Packet_Length > 0)													\
{																			\
	if (Packet_Length+Rdptr > BUFFER_SIZE+Rdbfr)							\
	{																		\
		if (Dry == false)													\
			fwrite(Rdptr, BUFFER_SIZE+Rdbfr-Rdptr, 1, dts[This_Track].file);\
		Packet_Length -= BUFFER_SIZE+Rdbfr-Rdptr;							\
		Refill_Buffer();													\
		Rdptr = Rdbfr;														\
	}																		\
	else																	\
	{																		\
		if (Dry == false)													\
			fwrite(Rdptr, Packet_Length, 1, dts[This_Track].file);			\
		Rdptr += Packet_Length;												\
		Packet_Length = 0;													\
	}																		\
}

static char *FTType[5] = {
	"48KHz", "44.1KHz", "44.1KHz", "44.1KHz", "44.1KHz"
};

static char *AC3ModeDash[8] = {
	"1+1", "1_0", "2_0", "3_0", "2_1", "3_1", "2_2", "3_2"
};

static char *AC3Mode[8] = {
	"1+1", "1/0", "2/0", "3/0", "2/1", "3/1", "2/2", "3/2"
};

static int AC3Rate[32] = {
	32, 40, 48, 56, 64, 80, 96, 112, 128, 160,
	192, 224, 256, 320, 384, 448, 512, 576, 640,
	0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0
};

// JEANL: Making this guys non-static so we can read them from elsewhere...
int VOB_ID, CELL_ID;
int NextVOB_ID, NextCELL_ID;
char Num_Buttons, NextNum_Buttons;
// VideoPTS is the i-frame PTS last read in the vob file. 
// CurrentDecodedFramePTS is the PTS of the current frame being sent to the store() function. Because
// store() might delay frames, it may not be the PTS of the frame currently displayed.
int VideoPTS, CurrentDecodedFramePTS, AudioPTS;	
static unsigned char PCM_Buffer[BUFFER_SIZE];
static short *ptrPCM_Buffer = (short*)PCM_Buffer;

int DrawButtons = 0;		// If 1, the button info is read, and the button rectangles drawn.
int ButtonsChanged = 1;		// Indicates button info did change since last data.
int StartVobNumber = 0;		// Number of first VOB in our array of files.
unsigned short NextStartY[36]; 
unsigned short NextEndY[36];	
unsigned short NextStartX[36];
unsigned short NextEndX[36];

unsigned short StartY[36]; 
unsigned short EndY[36];	
unsigned short StartX[36];
unsigned short EndX[36];

__int64 SeekFile(HANDLE hf, __int64 distance, DWORD MoveMethod);

int	PlayAudioStream;		// If not 0, stream ID of audio stream to playback.

bool FollowILVU;			// 0 to ignore ILVU (display them one after the other). 1 to follow ILVU pointers.	
int	ILVUJumpAtNextNavPack;	// If 0 no jump, otherwise indicates how many packs to skip to follow ILVU.
void Seek_Read_File(int Sectors);

bool AC3ChannelsFound[CHANNEL];
bool LPCMChannelsFound[CHANNEL];

void SelectAudioStream(int NewStreamID)
{
	int i;
	if(NewStreamID == PlayChannelID) return;

	// To switch audio track in real time, we need to reset the rip flag and the channel array.
	ZeroMemory(&Channel, sizeof(Channel));
	for (i=0; i<CHANNEL; i++)
	{
		ZeroMemory(&ac3[i], sizeof(AC3Stream));
		ZeroMemory(&mpa[i], sizeof(RAWStream));	
		ZeroMemory(&dts[i], sizeof(RAWStream));
	}

	ZeroMemory(&pcm, sizeof(struct PCMStream));
	pcm.rip = 0;
	PlayChannelID = PlayAudioStream = NewStreamID;
	FoundAudio = 0;
	NewAudioChannels = false;
	MYTRACE1("\nSwitching AUDIO %d\n",NewStreamID);
#if 0
if(NewStreamID>=SUB_AC3 && NewStreamID<SUB_AC3+CHANNEL)
	Channel[NewStreamID-SUB_AC3]=1, ac3[NewStreamID-SUB_AC3].rip=1;
if(NewStreamID>=SUB_PCM && NewStreamID<SUB_PCM+CHANNEL)
	Channel[NewStreamID-SUB_PCM]=1, pcm.rip=1;
#else
#endif
}

void Initialize_Buffer()
{
	BigBufferOffset = BIG_BUFFER_SIZE;
	BigBufferMaxOffset = BigBufferOffset;

	CurrentFileBytesRead = 0;
	CurrentFileSize = process.length[File_Flag];

	Rdptr = Rdbfr + BUFFER_SIZE;
	Rdmax = Rdptr;
	// This is important, so Next_Packet() doesn't return prematurely.
	NextBfr = 0;
	if (SystemStream_Flag)
	{
		if (Rdptr >= Rdmax)
			Next_Packet();
		CurrentBfr = *Rdptr++ << 24;

		if (Rdptr >= Rdmax)
			Next_Packet();
		CurrentBfr += *Rdptr++ << 16;

		if (Rdptr >= Rdmax)
			Next_Packet();
		CurrentBfr += *Rdptr++ << 8;

		if (Rdptr >= Rdmax)
			Next_Packet();
		CurrentBfr += *Rdptr++;

		Fill_Next();
	}
	else
	{
		Fill_Buffer();

		CurrentBfr = (*Rdptr << 24) + (*(Rdptr+1) << 16) + (*(Rdptr+2) << 8) + *(Rdptr+3);
		Rdptr += 4;

		Fill_Next();
	}

	BitsLeft = 32;
}

void Next_Packet()
{
	static int i, Packet_Length, Packet_Header_Length, size;
	static unsigned int code, AUDIO_ID, VOBCELL_Count, This_Track, MPA_Track;
	int ThisPTS;
	static int VobuCount = 0;

	for (;;)
	{
		// This is important if you're decoding a still video (1 I-frame, then audio) because there might not
		// be any more video frame to decode, where the Stop_Flag would be tested.
		if(Stop_Flag) break;
		code = Get_Short();
		code = (code<<16) + Get_Short();

		while ((code & 0xffffff00) != 0x00000100)
			code = (code<<8) + Get_Byte();

		switch (code)
		{
			case PACK_START_CODE:

				Rdptr += 8;
				VOBCELL_Count = 0;
				break;

			case PRIVATE_STREAM_2:
				Packet_Length = Get_Short();
//MYTRACE1("NavPack sector %d\n",_telli64(Infile[File_Flag]) / 2048 -1);
				if(ILVUJumpAtNextNavPack && (process.locate == LOCATE_FORWARD || process.locate == LOCATE_RIP))
				{
					// We're at a navpack where we should jump.
//					if(ILVUJumpAtNextNavPack == -1) ThreadKill();
//					MYTRACE1("Jump to start ILVU -> %d\n",_telli64(Infile[File_Flag]) / 2048 -1 + ILVUJumpAtNextNavPack-3);
					Rdptr = Rdbfr;
					Seek_Read_File(ILVUJumpAtNextNavPack-3);
					ILVUJumpAtNextNavPack = 0;
					break;
				}
				VobuCount ++;
				// If we've seen more than 2 vobus without the desired audio stream, let's declare there's no
				// audio...
				if(VobuCount > 2)
					FoundAudio = 0;
					
				if (++VOBCELL_Count==2)
				{
					short ILVUFlag = 0;
					int ILVUEnd = 0;
					int ILVUNext = 0;
					// JEANL: I modified this code because the VOB/CELL ID was inaccurate in some cases:
					// Sometimes, the pic was decoded OK, but then filling the next buffer caused the Nav
					// Pack to be parsed, and the VOB/CELL ID to be saved, even though they don't correspond
					// to the previous image. Here, I'm saving them as "Next" values. decode_image() will
					// update the fields and the true value based on these.
					Rdptr += 25;
					NextVOB_ID = Get_Short();
					// Was Get_Short(); But that was incorrect! The cellID is a byte, and the value next to it
					// is "reserved", 0 most of the time, but sometimes not!
					Rdptr ++;
					NextCELL_ID = Get_Byte();

					// Read ILVU information.
					Rdptr += 4;
					ILVUFlag = Get_Short();
					ILVUEnd = Get_Int();
					ILVUNext = Get_Int();
					ILVUJumpAtNextNavPack = 0;
					// If we want to follow ILVUs, check that ILVUs are present, and we're at the last VOBU.
					if(FollowILVU && (ILVUFlag & 0x4000) && (ILVUFlag & 0x1000))
					{
//						MYTRACE1("End of ILVU. Sector %d\n",_telli64(Infile[File_Flag]) / 2048 -1);
						if(ILVUNext > 0)
							ILVUJumpAtNextNavPack = ILVUNext - ILVUEnd + 1;
						else
							ILVUJumpAtNextNavPack = -1;
					}
					Rdptr += Packet_Length - 43;

					// Refresh audio streams.
					for(i=0;i<CHANNEL;i++)
					{
						if(FoundAC3Channels[i] != AC3ChannelsFound[i]) NewAudioChannels = true;
						if(FoundPCMChannels[i] != LPCMChannelsFound[i]) NewAudioChannels = true;
						FoundAC3Channels[i] = AC3ChannelsFound[i];
						FoundPCMChannels[i] = LPCMChannelsFound[i];
					}
					memset(AC3ChannelsFound,0,sizeof(AC3ChannelsFound));
					memset(LPCMChannelsFound,0,sizeof(LPCMChannelsFound));
				}
				else
				{
					// First PRIVATE_STREAM_2 packet.
					// Detect if we have a still frame.
					char LastNumButtons = NextNum_Buttons;
					int EndPTSIfEndOfSeq = 0;
					Rdptr += (0x41-0x2c);
					EndPTSIfEndOfSeq = Get_Int();
					if(EndPTSIfEndOfSeq) 
						StillFrame = 1;

					// Button count detection! 
					// The number of buttons is at offset 114 after the Length.
					Rdptr += 0x59 ; //(0x9e-0x45);
					NextNum_Buttons = Get_Byte();

#if 1
					// This code reads and parses the button info... They will be displayed in store.c
					// Note that I fill out the Next* arrays. The Current arrays are copied in Decode_Picture()
					// (in getpic.c).
					Rdptr += 0x1c;
					// ButtonsChanged will be reset by the function that draws the buttons.
					ButtonsChanged |= LastNumButtons != NextNum_Buttons;
//OutputDebugString("GetBit\n"); //Use this to see when this function is called.
					for(i=0;i<NextNum_Buttons;i++)
					{
						unsigned char Byte1 = Get_Byte();
						unsigned char Byte2 = Get_Byte();
						unsigned char Byte3 = Get_Byte();
						unsigned char Byte4 = Get_Byte();
						unsigned char Byte5 = Get_Byte();
						unsigned char Byte6 = Get_Byte();
						unsigned short LastVal;

						LastVal = NextStartX[i];
						NextStartX[i] = ((unsigned short)(Byte1 & 0x3f)) << 4; 
						NextStartX[i] |= ((unsigned short)(Byte2 & 0xf0)) >> 4;
						ButtonsChanged |= LastVal != NextStartX[i];
						LastVal = NextEndX[i];
						NextEndX[i] = ((unsigned short)(Byte2 & 0x03)) << 8; 
						NextEndX[i] |= (unsigned short)(Byte3);
						ButtonsChanged |= LastVal != NextEndX[i];

						LastVal = NextStartY[i];
						NextStartY[i] = ((unsigned short)(Byte4 & 0x3f)) << 4; 
						NextStartY[i] |= ((unsigned short)(Byte5 & 0xf0)) >> 4;
						ButtonsChanged |= LastVal != NextStartY[i];
						LastVal = NextEndY[i];
						NextEndY[i] = ((unsigned short)(Byte5 & 0x03)) << 8; 
						NextEndY[i] |= (unsigned short)(Byte6);
						ButtonsChanged |= LastVal != NextEndY[i];

						Rdptr+= 12;
					}
					// Reset point back to where we were right after reading NextNum_Buttons 
					Rdptr -= (0x1c + NextNum_Buttons*18);
#endif

					// Now reset pointer where we were, and resume the normal original code.
					Rdptr -= 115;
					Rdptr += Packet_Length;
				}
				break;

			case PRIVATE_STREAM_1:

				Packet_Length = Get_Short();

				Rdptr ++;	// +1
				code = Get_Byte();	// +1
				Packet_Header_Length = Get_Byte();	// +1

				// Jeanl: added to enable audio playback.
				if((process.locate == LOCATE_PLAY || process.locate == LOCATE_RIP) && PlayChannelID != 0)  
					Method_Flag = 2;

				if (code>=0x80)
				{
					code = Get_Byte();
					ThisPTS = (code & 0x0e) << 29;
					ThisPTS |= (Get_Short() & 0xfffe) << 14;
					ThisPTS |= (Get_Short()>>1) & 0x7fff;
					Rdptr += Packet_Header_Length-5;
				}
				else
				{
					// JEANL: If there's no PTS, an old value of ThisPTS will be displayed, that's
					// really bad when there's timing discontinuity. I'm setting this to 0.
					ThisPTS = 0;
					Rdptr += Packet_Header_Length;
				}

				AUDIO_ID = Get_Byte();	// +1
				Packet_Length -= Packet_Header_Length+4;

				if(AUDIO_ID >= 0x20 && AUDIO_ID < 0x40 && SubpicDecoder)
				{
					// This is a subtitle stream! Currently, we just skip it.
					// You'd call something like DEMUX_SUBPIC which would be quite similar to
					// DEMUX_AC3, and then send that to a class that can handle subpic decoding.
					DEMUX_SUBPIC
				}

				// Only set the audio PTS if we can really decode it...
				if((AUDIO_ID>=SUB_AC3 && AUDIO_ID<SUB_AC3+CHANNEL) || (AUDIO_ID>=SUB_PCM && AUDIO_ID<SUB_PCM+CHANNEL))
				{
//MYTRACE2("AudioPack sector %d ID %d\n",(int)_telli64(Infile[File_Flag]) / 2048 - 1, AUDIO_ID);
					AudioPTS = ThisPTS;
					if(AUDIO_ID>=SUB_AC3 && AUDIO_ID<SUB_AC3+CHANNEL)
						AC3ChannelsFound[AUDIO_ID-SUB_AC3] = 1;
					if(AUDIO_ID>=SUB_PCM && AUDIO_ID<SUB_PCM+CHANNEL)
						LPCMChannelsFound[AUDIO_ID-SUB_PCM] = 1;
				}

				// PlayChannelID == -2 indicates we should play the first audio stream
				// we encounter which we can play (AC3 or LPCM).
				if(PlayChannelID == -2 && ((AUDIO_ID>=SUB_AC3 && AUDIO_ID<SUB_AC3+CHANNEL) ||
				   (AUDIO_ID>=SUB_PCM && AUDIO_ID<SUB_PCM+CHANNEL))) 
				   PlayChannelID = AUDIO_ID;

				if (AUDIO_ID>=SUB_AC3 && AUDIO_ID<SUB_AC3+CHANNEL)
				{
					Rdptr += 3; Packet_Length -= 3; This_Track = AUDIO_ID-SUB_AC3;

					LOCATE

					if (!ac3[This_Track].rip && Start_Flag && !Channel[This_Track])
					{
						Channel[This_Track] = FORMAT_AC3;

						code = Get_Byte();
						code = (code & 0xff)<<8 | Get_Byte();
						i = 0;

						while (code!=0xb77)
						{
							code = (code & 0xff)<<8 | Get_Byte();
							i++;
						}

						Get_Short();
						ac3[This_Track].rate = (Get_Byte()>>1) & 0x1f;
						Get_Byte();
						ac3[This_Track].mode = (Get_Byte()>>5) & 0x07;

						Rdptr -= 7; Packet_Length -= i;

						if ((1 || D2V_Flag || Decision_Flag))
						{
							if (Decision_Flag && PlayChannelID == AUDIO_ID)
							{
								InitialAC3();

								DECODE_AC3

								ac3[This_Track].rip = true;
								Track_Flag = This_Track;
							}
							else if (Method_Flag==AUDIO_DECODE && PlayChannelID == AUDIO_ID)
							{
								sprintf(szBuffer, "%s AC3 T%02d %sch %dKbps %s.wav", szOutput, This_Track+1, 
									AC3ModeDash[ac3[This_Track].mode], AC3Rate[ac3[This_Track].rate], FTType[SRC_Flag]);

								strcpy(szBuffer, "C:\\temp\\audio.wav");
								strcpy(pcm.filename, szBuffer);
								// Jean: instead of saving the pcm data, let's send it to the audio output.
								FoundAudio = openWaveOut(48000, 2) != 0;
								VobuCount = 0;
								// This is questionable. The VideoPTS we have here is the I-Frame PTS, and
								// it's not guaranteed to the the PTS of the first displayed frame.
								// In most cases, it isn't.
								pcm.delay = ((AudioPTS-VideoPTS)/90) * 192;
								if (SRC_Flag)
								{
									InitialSRC();
								}

								if (pcm.delay > 0)
								{
									if (SRC_Flag)
										pcm.delay = ((int)(0.91875*pcm.delay)>>2)<<2;

									pcm.size += pcm.delay;
									pcm.delay = 0;
								}

								InitialAC3();

								DECODE_AC3

								if (-pcm.delay > size)
									pcm.delay += size;
								else
								{
									WriteOut(AC3Dec_Buffer-pcm.delay, size+pcm.delay, AudioPTS-pcm.delay/2/2/48*90);
									pcm.size += size+pcm.delay;
 									pcm.delay = 0;
								}
								Track_Flag = This_Track;
								ac3[This_Track].rip = true;
//								FoundAudio = true;
								VobuCount = 0;

							}
							else if (Method_Flag==AUDIO_DEMUXALL  ||  Method_Flag==AUDIO_DEMUX && PlayChannelID == AUDIO_ID)
							{
								sprintf(szBuffer, "%s AC3 T%02d %sch %dKbps DELAY %dms.ac3", szOutput, This_Track+1, 
									AC3ModeDash[ac3[This_Track].mode], AC3Rate[ac3[This_Track].rate], (AudioPTS-VideoPTS)/90);
							}
						}
					}
					else if (ac3[This_Track].rip)
					{
						if (Decision_Flag)
							DECODE_AC3
						else if (Method_Flag==AUDIO_DECODE)
						{
							DECODE_AC3

							if (-pcm.delay > size)
								pcm.delay += size;
							else
							{
								if (SRC_Flag)
								{
								}
								else if (Dry == false)
								{
									// Send the samples to the audio output... Modify Audio PTS to reflect how many samples
									// we're skipping.
//									FoundAudio = true;
									VobuCount = 0;
									WriteOut(AC3Dec_Buffer-pcm.delay, size+pcm.delay, AudioPTS -pcm.delay/2/2/48*90);
								}

								if (Dry == false)
									pcm.size += size+pcm.delay;
 								pcm.delay = 0;
							}
						}
						else
							DEMUX_AC3
					}
				}
				else if (AUDIO_ID>=SUB_PCM && AUDIO_ID<SUB_PCM+CHANNEL)
				{
					Rdptr += 6; Packet_Length -= 6; This_Track = AUDIO_ID-SUB_PCM;

					LOCATE

					if (!pcm.rip && Start_Flag && !Channel[This_Track])
					{
						Channel[This_Track] = FORMAT_LPCM;

						if (1 || AVI_Flag || D2V_Flag)
						{
							if (Method_Flag==AUDIO_DECODE && PlayChannelID == AUDIO_ID)
							{
								pcm.delay = ((AudioPTS-VideoPTS)/90) * 192;

								sprintf(szBuffer, "%s LPCM T%02d %s.wav", szOutput, This_Track+1, FTType[SRC_Flag]);
								strcpy(pcm.filename, szBuffer);

								FoundAudio = openWaveOut(48000, 2);
								VobuCount = 0;

								if (SRC_Flag)
								{
									InitialSRC();
								}

								if (pcm.delay > 0)
								{
									if (SRC_Flag)
										pcm.delay = ((int)(0.91875*pcm.delay)>>2)<<2;

									pcm.size += pcm.delay;
									pcm.delay = 0;
								}

								if (-pcm.delay > Packet_Length)
									pcm.delay += Packet_Length;
								else
								{
									DEMUX_PCM

									WriteOut(PCM_Buffer-pcm.delay, size+pcm.delay, AudioPTS-pcm.delay/2/2/48*90);
									pcm.size += size+pcm.delay;
									pcm.delay = 0;
								}
								Track_Flag = This_Track;
								pcm.rip = true;
//								FoundAudio = true;
								VobuCount = 0;
							}
						}
					}
					else if (pcm.rip)
					{
						if (-pcm.delay > Packet_Length)
							pcm.delay += Packet_Length;
						else
						{
							DEMUX_PCM

							if (SRC_Flag)
							{
							}
							else
							{
								WriteOut(PCM_Buffer-pcm.delay, size+pcm.delay, AudioPTS-pcm.delay/2/2/48*90);
//								FoundAudio = true;
								VobuCount = 0;
							
								if (Norm_Flag)
									for (i=0; i<(size>>1); i++)
										if (Sound_Max < abs(ptrPCM_Buffer[i]))
											Sound_Max = abs(ptrPCM_Buffer[i]);
							}

							if (Dry == false)
								pcm.size += size+pcm.delay;
							pcm.delay = 0;
						}
					}
				}
				else if (AUDIO_ID>=SUB_DTS && AUDIO_ID<SUB_DTS+CHANNEL)
				{
					Rdptr += 3; Packet_Length -= 3; This_Track = AUDIO_ID-SUB_DTS;

					LOCATE

					if (!dts[This_Track].rip && Start_Flag && !Channel[This_Track])
					{
						Channel[This_Track] = FORMAT_DTS;

						if (AVI_Flag || D2V_Flag)
						{
							if (Method_Flag==AUDIO_DEMUXALL  ||  Method_Flag==AUDIO_DEMUX && PlayChannelID == AUDIO_ID)
							{
								sprintf(szBuffer, "%s DTS T%02d DELAY %dms.dts", szOutput, This_Track+1, (AudioPTS-VideoPTS)/90);

								dts[This_Track].file = fopen(szBuffer, "wb");

								DEMUX_DTS

								dts[This_Track].rip = true;
		 					}
						}
					}
					else if (dts[This_Track].rip)
						DEMUX_DTS
				}
				Rdptr += Packet_Length;
				break;

			case AUDIO_ELEMENTARY_STREAM_7:
				MPA_Track++;
			case AUDIO_ELEMENTARY_STREAM_6:
				MPA_Track++;
			case AUDIO_ELEMENTARY_STREAM_5:
				MPA_Track++;
			case AUDIO_ELEMENTARY_STREAM_4:
				MPA_Track++;
			case AUDIO_ELEMENTARY_STREAM_3:
				MPA_Track++;
			case AUDIO_ELEMENTARY_STREAM_2:
				MPA_Track++;
			case AUDIO_ELEMENTARY_STREAM_1:
				MPA_Track++;
			case AUDIO_ELEMENTARY_STREAM_0:
				Packet_Length = Get_Short()-1;
				code = Get_Byte();

#if 0
				// Jean: remove mpeg audio demuxing.
				if ((code & 0xc0)==0x80)
				{
					code = Get_Byte();	// +1
					Packet_Header_Length = Get_Byte();	// +1

					if (code>=0x80)
					{
						code = Get_Byte();

						AudioPTS = (code & 0x0e) << 29;
						AudioPTS |= (Get_Short() & 0xfffe) << 14;
						AudioPTS |= (Get_Short()>>1) & 0x7fff;

						Rdptr += Packet_Header_Length-5;
					}
					else
						Rdptr += Packet_Header_Length;

					Packet_Length -= Packet_Header_Length+2;

					LOCATE

					if (!mpa[MPA_Track].rip && Start_Flag && !Channel[MPA_Track])
					{
						Channel[MPA_Track] = FORMAT_MPA;

						code = Get_Byte();
						code = (code & 0xff)<<8 | Get_Byte();
						i = 0;

						while (code<0xfff0)
						{
							code = (code & 0xff)<<8 | Get_Byte();
							i++;
						}

						Rdptr -= 2; Packet_Length -= i;

						if ((1 || D2V_Flag))
						{
							if (Method_Flag==AUDIO_DEMUXALL  ||  Method_Flag==AUDIO_DEMUX && MPA_Track==Track_Flag)
							{
								sprintf(szBuffer, "%s MPA T%02d DELAY %dms.mpa", szOutput, MPA_Track+1, (AudioPTS-VideoPTS)/90);
								mpa[MPA_Track].file = fopen(szBuffer, "wb");

								DEMUX_MPA

								mpa[MPA_Track].rip = true;
							}
						}
					}
					else if (mpa[MPA_Track].rip)
						DEMUX_MPA
				}
#endif

				Rdptr += Packet_Length;
				MPA_Track = 0;
				break;

			case VIDEO_ELEMENTARY_STREAM:
				Packet_Length = Get_Short();
				Rdmax = Rdptr + Packet_Length;

				code = Get_Byte();

				if ((code & 0xc0)==0x80)
				{
					code = Get_Byte();
					Packet_Header_Length = Get_Byte();
//jsoto Jan-2007
					if (code>=0x80)
					{
						code = Get_Byte();
						VideoPTS = (code & 0x0e) << 29;
						VideoPTS |= (Get_Short() & 0xfffe) << 14;
						VideoPTS |= (Get_Short()>>1) & 0x7fff;
						// Adjust VideoPTS as required if any B frames precede the first I frame
						// in display order.
						if (Frame_Rate && !Start_Flag)
							VideoPTS -= (int) ((LeadingBFrames * 90000) / Frame_Rate);

						Rdptr += Packet_Header_Length-5;
					}
					else
						Rdptr += Packet_Header_Length;

					Bitrate_Monitor += Rdmax-Rdptr;
//MYTRACE1("Video sector %d\n",_telli64(Infile[File_Flag]) / 2048 - 1);
					return;
				}
				else
				{
					Rdptr += Packet_Length-1;
				}

				break;

			default:
				if (code>=SYSTEM_START_CODE)
				{
					Packet_Length = Get_Short();
					Rdptr += Packet_Length;
				}
				break;
		}
	}
}

unsigned int Get_Bits_All(unsigned int N)
{
	N -= BitsLeft;
	Val = (CurrentBfr << (32 - BitsLeft)) >> (32 - BitsLeft);

	if (N)
		Val = (Val << N) + (NextBfr >> (32 - N));

	CurrentBfr = NextBfr;
	BitsLeft = 32 - N;
	Fill_Next();

	return Val;
}

void Flush_Buffer_All(unsigned int N)
{
	CurrentBfr = NextBfr;
	BitsLeft = BitsLeft + 32 - N;


	// Code to test whether we're at a sequence_end_code, in which case we should return without calling
	// Fill_Next(). 
	if((NextBfr == SEQUENCE_END_CODE) ||
		((NextBfr & 0x00FFFFFF) == 0x01 && Rdptr[0] == 0xB7) || 
		((NextBfr & 0x0000FFFF) == 0 && Rdptr[0] == 0x01 && Rdptr[1] == 0xB7) ||
		((NextBfr & 0x000000FF) == 0 && Rdptr[0] == 0x00 && Rdptr[1] == 0x01 && Rdptr[2] == 0xB7)
		)
	{
		// Set StillFrame flag, but don't try to fill NextBfr, because that might trigger a search for the next video.
		Rdptr += 4;
		StillFrame = 1;
	}
	else
		Fill_Next();
}


void Refill_Buffer(void)
{
	if(BigBufferOffset >= BIG_BUFFER_SIZE)
	{
		// We need more from the file!
		ZeroMemory(&BigBuffer, BIG_BUFFER_SIZE);

		CurrentFileBytesRead = SeekFile(Infile[File_Flag], 0, FILE_CURRENT);

		ReadFile(Infile[File_Flag], &BigBuffer[0], BIG_BUFFER_SIZE, &NumberOfBytesRead, NULL);

		if (NumberOfBytesRead < BIG_BUFFER_SIZE)
			Next_File();

		BigBufferOffset = 0;
		BigBufferMaxOffset = NumberOfBytesRead;
	}

	if(BigBufferOffset >= BigBufferMaxOffset)
	{
		// Hmm... we're totally out of data!
		ZeroMemory(&BigBuffer, BIG_BUFFER_SIZE);

		BigBufferOffset = 0;
		BigBufferMaxOffset = BUFFER_SIZE;

		// We can't do a thread kill here because we'll lose
		// the last two frames. Set up to kill at the next Get_Hdr()
		// call. We allow the last file to wrap around but we won't
		// use any of it after wrapping. I know, it's kludgy, but it works.
		if(Dry == true)
		{
			// JEANL: I'm not sure I understand the meaning of the Dry variable, but it seems
			// that if we try to open an invalid VOB file (encrypted for example) or a file of
			// some other type, we end up here. We don't if the vob files are valid.
			Fault_Flag = CRITICAL_ERROR_LEVEL;
			// I'm adding a sector hit here, because it's possible we'll reach the end of the vob
			// and not detect we reached the target sector...
			SectorHit = 1;
			ThreadKill();
		}
		Dry = true;
	}

	CopyMemory(&Rdbfr[0], &BigBuffer[BigBufferOffset], BUFFER_SIZE);

	BigBufferOffset += BUFFER_SIZE;
	
	if(Dry == false)
	{
		if(CurrentFileBytesRead == CurrentFileSize)
		{
			CurrentFileBytesRead = 0;
			CurrentFileSize = process.length[File_Flag];
		}

		CurrentFileBytesRead += BUFFER_SIZE;
	}
}

void Fill_Buffer()
{
	Refill_Buffer();

	if (KeyOp_Flag && (Rdbfr[20] & 0x10))
	{
		BufferOp(Rdbfr, lfsr0, lfsr1);
		Rdbfr[20] &= ~0x10;
	}

	Rdptr = Rdbfr;

	if (SystemStream_Flag)
		Rdmax -= BUFFER_SIZE;
	else
		Bitrate_Monitor += NumberOfBytesRead;
}

void Next_File()
{
	int i;
	DWORD TempNumberOfBytesRead = 0;
	bool AddNumberOfBytesRead = false;

	process.run = 0;
	for (i = 0; i <= File_Flag; i++)
	{
		process.run += process.length[i];
	}

	if (File_Flag < File_Limit-1)
	{
		File_Flag++;

		SeekFile(Infile[File_Flag], 0, FILE_BEGIN);
		ReadFile(Infile[File_Flag], &BigBuffer[NumberOfBytesRead], BIG_BUFFER_SIZE - NumberOfBytesRead, &TempNumberOfBytesRead, NULL);

		NumberOfBytesRead += TempNumberOfBytesRead;
	}
}

void Seek_Read_File(int Sectors)
{
	__int64 CurrentLoc;

	Rdptr = Rdbfr;

	CurrentLoc = SeekFile(Infile[File_Flag], 0, FILE_CURRENT);
	if(CurrentLoc == -1)
	{
		ThreadKill();
	}
	else
	{
		// To keep things simple, we seek backwards in the file(s) until we're at the point that
		// represents the current sector being played and not where we've got up to with the buffering.
		__int64 OffsetLoc;

		OffsetLoc = (BigBufferMaxOffset - BigBufferOffset);

		if(CurrentLoc >= OffsetLoc)
		{
			// The location we require is in this file.

			// Seek backwards to that location.
			if(SeekFile(Infile[File_Flag], -OffsetLoc, FILE_CURRENT) == -1)
			{
				// Doh!
				ThreadKill();
				return;
			}
		}
		else
		{
		    // The location we require is actually in the previous file.

			// Seek to the start of this file - just for good measure!
			if(SeekFile(Infile[File_Flag], 0, FILE_BEGIN) == -1)
			{
				// Doh!
				ThreadKill();
				return;
			}

			// Recalculate the distance we need to seek.
			OffsetLoc -= CurrentLoc;
			
			// Change to the previous file.
			File_Flag--;
			CurrentFileBytesRead = process.length[File_Flag];
			CurrentFileSize = process.length[File_Flag];

			// Seek backwards from the end of the file.
			if(SeekFile(Infile[File_Flag], -OffsetLoc, FILE_END) == -1)
			{
				// Doh!
				ThreadKill();
				return;
			}
		}

		// Now we can continue with the original seeks.
		if(SeekFile(Infile[File_Flag], (Sectors * BUFFER_SIZE), FILE_CURRENT) != -1)
		{
			BigBufferOffset = BIG_BUFFER_SIZE; // Force a read from the file
            Refill_Buffer();
		}
		else if(File_Flag < File_Limit-1)
		{
			int CurSector = (int)(SeekFile(Infile[File_Flag], 0, FILE_CURRENT) / BUFFER_SIZE);
			Sectors -= (int)((process.length[File_Flag] / BUFFER_SIZE) - CurSector);
			File_Flag++;
			CurrentFileBytesRead = 0;
			CurrentFileSize = process.length[File_Flag];

			if(SeekFile(Infile[File_Flag], (Sectors * BUFFER_SIZE), FILE_BEGIN) == -1)
			{
				BigBufferOffset = BIG_BUFFER_SIZE; // Force a read from the file
                Refill_Buffer();
			}
			else
			{
				ThreadKill();
			}
		}
		else
		{
			ThreadKill();
		}
	}
}

void UpdateInfo()
{
	static int Old_NTSC_Purity;
	int pts;

	sprintf(szBuffer, "%s", AspectRatio[aspect_ratio_information]);
	SetDlgItemText(hDlg, IDC_ASPECT_RATIO, szBuffer);

	sprintf(szBuffer, "%.3f fps", Frame_Rate);
	SetDlgItemText(hDlg, IDC_FRAME_RATE, szBuffer);

	if(process.locate==LOCATE_RIP && PlayChannelID != 0 && PlayChannelID != -1 && FoundAudio)
	{
		switch (Channel[Track_Flag])
		{
			case FORMAT_AC3:
				sprintf(szBuffer, "AC3 %s %d", AC3Mode[ac3[Track_Flag].mode], AC3Rate[ac3[Track_Flag].rate]);
				SetDlgItemText(hDlg, IDC_AUDIO_TYPE, szBuffer);
				sprintf(szBuffer, "0x%x",PlayChannelID);
				SetDlgItemText(hDlg, IDC_AUDIOCHANNEL, szBuffer);
				break;

			case FORMAT_MPA:
				SetDlgItemText(hDlg, IDC_AUDIO_TYPE, "MPEG Audio (can't play)");
				break;

			case FORMAT_LPCM:
				SetDlgItemText(hDlg, IDC_AUDIO_TYPE, "Linear PCM");
				sprintf(szBuffer, "0x%x",PlayChannelID);
				SetDlgItemText(hDlg, IDC_AUDIOCHANNEL, szBuffer);
				break;

			case FORMAT_DTS:
				SetDlgItemText(hDlg, IDC_AUDIO_TYPE, "DTS (can't play)");
				break;

			default:
				SetDlgItemText(hDlg, IDC_AUDIO_TYPE, "");
				break;
		}
	}
	else
	{
		SetDlgItemText(hDlg, IDC_AUDIO_TYPE, "");
		SetDlgItemText(hDlg, IDC_AUDIOCHANNEL, "");
	}

	if (SystemStream_Flag && process.locate!=LOCATE_INIT)
	{
// jsoto Jan-2007
//		pts = AudioPTS/90000;
//		sprintf(szBuffer, "%d:%02d:%02d", pts/3600, (pts%3600)/60, pts%60);
//		SetDlgItemText(hDlg, IDC_TIMESTAMP, szBuffer);
		pts = VideoPTS/90000;
//		sprintf(szBuffer, "%d:%02d:%02d.%02d", pts/3600, (pts%3600)/60, pts%60,(VideoPTS/900)%100);
		sprintf(szBuffer, "%d:%02d:%02d.%02d", pts/3600, (pts%3600)/60, pts%60,(VideoPTS % 90000) / 
			((frame_rate <27) ? 3600 : 3003));
		SetDlgItemText(hDlg, IDC_TIMESTAMP, szBuffer);
	}
	else
		SetDlgItemText(hDlg, IDC_TIMESTAMP, "");

	if(d2v_current.file+StartVobNumber == -1)
	{
		sprintf(szBuffer,"%d / -",(int)d2v_current.lba);
	}
	else
	{
		sprintf(szBuffer,"%d / %d",(int)d2v_current.lba,d2v_current.file+StartVobNumber);
	}
	SetDlgItemText(hDlg, IDC_SECTOR,szBuffer);

	if (process.locate==LOCATE_RIP)
	{
		if (NTSC_Purity || FILM_Purity)
		{
			if (!FILM_Purity)
			{
				if (frame_rate==25)
					sprintf(szBuffer, "PAL");
				else
					sprintf(szBuffer, "NTSC");
			}
			else if (!NTSC_Purity)
				sprintf(szBuffer, "FILM");
			else if (NTSC_Purity > Old_NTSC_Purity)
				sprintf(szBuffer, "NTSC %2d %%", 100 - (FILM_Purity*100)/(FILM_Purity+NTSC_Purity));
			else
				sprintf(szBuffer, "FILM %2d %%", (FILM_Purity*100)/(FILM_Purity+NTSC_Purity));

			Old_NTSC_Purity = NTSC_Purity;
			SetDlgItemText(hDlg, IDC_VIDEO_TYPE, szBuffer);
		}

		if (process.op)
		{
			__int64 CurrentLoc;
			
//			CurrentLoc = SeekFile(Infile[File_Flag], 0, FILE_CURRENT) - (BigBufferMaxOffset - BigBufferOffset);
            CurrentLoc = CurrentFileBytesRead;

			process.ed = timeGetTime();
			process.elapsed = (process.ed-process.op)/1000;
			process.percent = (float)(100.0*(process.run-process.start+CurrentLoc)/(process.end-process.start));
			process.remain = (int)((process.ed-process.op)*(100.0-process.percent)/process.percent)/1000;

			sprintf(szBuffer, "%d:%02d:%02d", process.elapsed/3600, (process.elapsed%3600)/60, process.elapsed%60);
			SetDlgItemText(hDlg, IDC_ELAPSED, szBuffer);

			sprintf(szBuffer, "%d:%02d:%02d", process.remain/3600, (process.remain%3600)/60, process.remain%60);
			SetDlgItemText(hDlg, IDC_REMAIN, szBuffer);
		}
	}
	else if (GetDlgItemText(hDlg, IDC_ELAPSED, szBuffer, 9))
	{
		SetDlgItemText(hDlg, IDC_VIDEO_TYPE, "");
		SetDlgItemText(hDlg, IDC_FRAME_TYPE, "");
		SetDlgItemText(hDlg, IDC_CODED_NUMBER, "");
		SetDlgItemText(hDlg, IDC_PLAYBACK_NUMBER, "");
		SendDlgItemMessage(hDlg, IDC_PROGRESS, PBM_SETPOS, 0, 0);
		SetDlgItemText(hDlg, IDC_FILE, "");
		SetDlgItemText(hDlg, IDC_FILE_SIZE, "");
		SetDlgItemText(hDlg, IDC_ELAPSED, "");
		SetDlgItemText(hDlg, IDC_REMAIN, "");
		SetDlgItemText(hDlg, IDC_FPS, "");
	}
}
